# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests service configuration."""

from invenio_records_resources.services import RecordServiceConfig, SearchOptions
from invenio_records_resources.services.records.links import pagination_links

from invenio_requests.services.requests import facets

from ...customizations import RequestActions
from ...records.api import Request
from ..configurator import ConfiguratorMixin, FromConfig
from ..permissions import PermissionPolicy
from .components import (
    EntityReferencesComponent,
    RequestDataComponent,
    RequestNumberComponent,
)
from .links import RequestLink
from .params import IsOpenParam, ReferenceFilterParam
from .results import RequestItem, RequestList


def _is_action_available(request, context):
    """Check if the given action is available on the request."""
    action = context.get("action")
    identity = context.get("identity")
    permission_policy_cls = context.get("permission_policy_cls")
    permission = permission_policy_cls(f"action_{action}", request=request)
    return RequestActions.can_execute(request, action) and permission.allows(identity)


class RequestSearchOptions(SearchOptions):
    """Search options."""

    params_interpreters_cls = SearchOptions.params_interpreters_cls + [
        ReferenceFilterParam.factory(param="created_by", field="created_by"),
        ReferenceFilterParam.factory(param="receiver", field="receiver"),
        ReferenceFilterParam.factory(param="topic", field="topic"),
        IsOpenParam.factory("is_open"),
    ]

    facets = {
        "type": facets.type,
        "status": facets.status,
    }


class RequestsServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Requests service configuration."""

    # common configuration
    permission_policy_cls = FromConfig(
        "REQUESTS_PERMISSION_POLICY", default=PermissionPolicy
    )
    result_item_cls = RequestItem
    result_list_cls = RequestList
    search = RequestSearchOptions

    # request-specific configuration
    record_cls = Request  # needed for model queries
    schema = None  # stored in the API classes, for customization
    index_dumper = None

    # links configuration
    links_item = {
        "self": RequestLink("{+api}/requests/{id}"),
        "comments": RequestLink("{+api}/requests/{id}/comments"),
        "timeline": RequestLink("{+api}/requests/{id}/timeline"),
    }
    links_search = pagination_links("{+api}/requests{?args*}")
    links_user_requests_search = pagination_links("{+api}/user/requests{?args*}")
    action_link = RequestLink(
        "{+api}/requests/{id}/actions/{action}", when=_is_action_available
    )

    components = [
        # Order of components are important!
        RequestDataComponent,
        EntityReferencesComponent,
        RequestNumberComponent,
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resolver and proxy for requests."""

from invenio_records_resources.references.resolvers.base import (
    EntityProxy,
    EntityResolver,
)
from sqlalchemy.exc import StatementError
from sqlalchemy.orm.exc import NoResultFound

from ..records.api import Request


class RequestProxy(EntityProxy):
    """Resolver proxy for a Request entity."""

    def _resolve(self):
        """Resolve the Request from the proxy's reference dict."""
        request_id = self._parse_ref_dict_id()
        try:
            return Request.get_record(request_id)
        except StatementError as exc:
            raise NoResultFound() from exc

    def get_need(self):
        """Return None since Needs are not applicable to requests."""
        return None


class RequestResolver(EntityResolver):
    """Resolver for requests."""

    type_id = "request"

    def matches_entity(self, entity):
        """Check if the entity is a request."""
        return isinstance(entity, Request)

    def _reference_entity(self, entity):
        """Create a reference dict for the given request."""
        id_ = entity.number if entity.number is not None else entity.id
        return {"request": str(id_)}

    def matches_reference_dict(self, ref_dict):
        """Check if the reference dict references a request."""
        return self._parse_ref_dict_type(ref_dict) == "request"

    def _get_entity_proxy(self, ref_dict):
        """Return a RequestProxy for the given reference dict."""
        return RequestProxy(ref_dict)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfields for request records."""

from .entity_reference import EntityReferenceField
from .event_type import EventTypeField
from .expired_state import ExpiredStateCalculatedField
from .identity import IdentityField
from .request_state import RequestStateCalculatedField
from .request_type import RequestTypeField
from .status import RequestStatusField

__all__ = (
    "EntityReferenceField",
    "EventTypeField",
    "ExpiredStateCalculatedField",
    "IdentityField",
    "RequestStateCalculatedField",
    "RequestStatusField",
    "RequestTypeField",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base class for creating custom types of requests.

The `RequestType` classes are the most important part in the customization/extension
mechanism for custom types of requests.
TODO explain what can be done here, and how!
"""


import base32_lib as base32
import marshmallow as ma
from invenio_records_resources.services.references import EntityReferenceBaseSchema

from ..proxies import current_requests
from .actions import (
    AcceptAction,
    CancelAction,
    CreateAction,
    DeclineAction,
    DeleteAction,
    ExpireAction,
    SubmitAction,
)
from .states import RequestState


class RequestType:
    """Base class for custom request types."""

    type_id = "base-request"
    """The unique and constant identifier for this type of requests.

    Since this property is used to map generic chunks of data from the database
    (i.e. the request model entries) to their correct `RequestType`, this should
    be a globally unique value.
    By convention, this would be the name of the package in which the custom
    `RequestType` is defined as prefix, together with a suffix related to the
    `RequestType`.
    Further, it should be constant after the first release of the package
    (otherwise, requests created with the old value will no longer be able to be
    mapped to their `RequestType`).
    """

    name = "Base Request"
    """The human-readable name for this type of requests."""

    available_statuses = {
        "created": RequestState.UNDEFINED,
        "submitted": RequestState.OPEN,
        "cancelled": RequestState.CLOSED,
        "declined": RequestState.CLOSED,
        "accepted": RequestState.CLOSED,
        "expired": RequestState.CLOSED,
        "deleted": RequestState.UNDEFINED,
    }
    """Available statuses for the request.

    The keys in this dictionary is the set of available statuses, and their
    values are indicators whether this request is considered to be open, closed
    or undefined.
    """

    create_action = "create"
    """Defines the action that's able to create this request.

    This must be set to one of the available actions for the custom request type.
    """

    delete_action = "delete"
    """Defines the action that's able to create this request.

    This must be set to one of the available actions for the custom request type.
    """

    available_actions = {
        "create": CreateAction,
        "submit": SubmitAction,
        "delete": DeleteAction,
        "accept": AcceptAction,
        "cancel": CancelAction,
        "decline": DeclineAction,
        "expire": ExpireAction,
    }
    """Available actions for this Request.

    The keys are the internal identifiers for the actions, the values are
    the actual RequestAction classes (not objects).
    Whenever an action is looked up, a new object of the registered
    RequestAction class is instantiated with the current Request object as
    argument.
    """

    creator_can_be_none = True
    """Determines if the ``created_by`` reference accepts ``None``."""

    receiver_can_be_none = False
    """Determines if the ``receiver`` reference accepts ``None``."""

    topic_can_be_none = True
    """Determines if the ``topic`` reference accepts ``None``."""

    allowed_creator_ref_types = ["user"]
    """A list of allowed TYPE keys for ``created_by`` reference dicts."""

    allowed_receiver_ref_types = ["user"]
    """A list of allowed TYPE keys for ``receiver`` reference dicts."""

    allowed_topic_ref_types = []
    """A list of allowed TYPE keys for ``topic`` reference dicts."""

    payload_schema = None
    """Schema for supported payload fields.

    Define it as a dictionary of fields mapping:

    .. code-block:: python

        payload_schema = {
            "content": fields.String(),
            # ...
        }
    """

    needs_context = None
    """The context for needs.

    This allows a request type to customize which and how needs are generated
    for a specific entity. For instance which needs are generated for a
    community entity or system entity.
    """

    @classmethod
    def entity_needs(cls, entity):
        """Generate entity needs for the given entity."""
        if entity is not None:
            return entity.get_needs(ctx=cls.needs_context)
        return []

    @classmethod
    def _create_marshmallow_schema(cls):
        """Create a marshmallow schema for this request type."""
        # Avoid circular imports
        from invenio_requests.services.schemas import RequestSchema

        # The reference fields always need to be added
        RefBaseSchema = EntityReferenceBaseSchema
        additional_fields = {
            "created_by": ma.fields.Nested(
                RefBaseSchema.create_from_dict(cls.allowed_creator_ref_types),
                allow_none=cls.creator_can_be_none,
            ),
            "receiver": ma.fields.Nested(
                RefBaseSchema.create_from_dict(cls.allowed_receiver_ref_types),
                allow_none=cls.receiver_can_be_none,
            ),
            "topic": ma.fields.Nested(
                RefBaseSchema.create_from_dict(cls.allowed_topic_ref_types),
                allow_none=cls.topic_can_be_none,
            ),
        }

        # Raise on invalid payload keys
        class PayloadBaseSchema(ma.Schema):
            class Meta:
                unknown = ma.RAISE

        # If a payload schema is defined, add it to the request schema
        if cls.payload_schema is not None:
            additional_fields["payload"] = ma.fields.Nested(
                PayloadBaseSchema.from_dict(cls.payload_schema),
            )

        # Dynamically create a schema from the fields defined
        # by the payload schema dict.
        return RequestSchema.from_dict(additional_fields)

    @classmethod
    def marshmallow_schema(cls):
        """Create a schema for the entire request including payload."""
        type_id = cls.type_id
        if type_id not in current_requests._schema_cache:
            current_requests._schema_cache[type_id] = cls._create_marshmallow_schema()
        return current_requests._schema_cache[type_id]

    def generate_request_number(self, request, **kwargs):
        """Generate a new request number identifier.

        This method can be overridden in subclasses to create external identifiers
        according to a custom schema, using the information associated with the request
        (e.g. topic, receiver, creator).
        """
        from invenio_requests.records.models import RequestNumber

        return base32.encode(RequestNumber.next())

    def __str__(self):
        """Return str(self)."""
        # Value used by marshmallow schemas to represent the type.
        return self.type_id

    def __repr__(self):
        """Return repr(self)."""
        return f"<RequestType '{self.type_id}'>"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Utility for rendering URI template links."""

from invenio_records_resources.services.base.links import Link, LinksTemplate


class RequestLinksTemplate(LinksTemplate):
    """Templates for generating links for a request object."""

    def __init__(self, links, action_link, context=None):
        """Constructor."""
        super().__init__(links, context=context)
        self._action_link = action_link

    def expand(self, req, identity=None):
        """Expand all the link templates."""
        links = {}

        # expand links for all available actions on the request
        links["actions"] = {}
        link = self._action_link
        for action in req.type.available_actions:
            if action in [req.type.create_action, req.type.delete_action]:
                continue
            ctx = self.context.copy()
            ctx["action"] = action
            ctx["identity"] = identity
            if link.should_render(req, ctx):
                links["actions"][action] = link.expand(req, ctx)

        # expand the other configured links
        for key, link in self._links.items():
            if link.should_render(req, self.context):
                links[key] = link.expand(req, self.context)

        return links


class RequestLink(Link):
    """Shortcut for writing request links."""

    @staticmethod
    def vars(record, vars):
        """Variables for the URI template."""
        # TODO this uses the UUID of the record, should we maybe use the number/ext-id?
        vars.update({"id": record.id})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helpers for customizing the configuration in a controlled manner."""

from invenio_base.utils import load_or_import_from_config


class FromConfig:
    """Data descriptor to connect config with application configuration.

    See https://docs.python.org/3/howto/descriptor.html .

    .. code-block:: python

        # service/config.py
        class ServiceConfig:
            foo = FromConfig("FOO", default=1)

        # config.py
        FOO = 2

        # ext.py
        c = ServiceConfig.build(app)
        c.foo  # 2
    """

    def __init__(self, config_key, default=None):
        """Constructor for data descriptor."""
        self.config_key = config_key
        self.default = default

    def __get__(self, obj, objtype=None):
        """Return value that was grafted on obj (descriptor protocol)."""
        return load_or_import_from_config(
            app=obj._app, key=self.config_key, default=self.default
        )

    def __set_name__(self, owner, name):
        """Store name of grafted field (descriptor protocol)."""
        # If we want to allow setting it we can implement this.
        pass

    def __set__(self, obj, value):
        """Set value on grafted_field of obj (descriptor protocol)."""
        # If we want to allow setting it we can implement this.
        pass


class ConfiguratorMixin:
    """Shared customization for requests service config."""

    @classmethod
    def build(cls, app):
        """Build the config object."""
        return type(f"Custom{cls.__name__}", (cls,), {"_app": app})()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resolver and proxy for requests."""

from invenio_records_resources.references.registry import ResolverRegistryBase

from ..proxies import current_requests


class ResolverRegistry(ResolverRegistryBase):
    """Namespace for resolver functions."""

    @classmethod
    def get_registered_resolvers(cls):
        """Get all currently registered resolvers."""
        return iter(current_requests.entity_resolvers_registry)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfield for calculating the ``is_expired`` property of a request."""


from datetime import datetime

import pytz
from invenio_records_resources.records.systemfields.calculated import CalculatedField


class ExpiredStateCalculatedField(CalculatedField):
    """Systemfield for calculating whether or not the request is expired."""

    def __init__(self, key=None):
        """Constructor."""
        super().__init__(key=key, use_cache=False)

    def calculate(self, record):
        """Calculate the ``is_expired`` property of the request."""
        expires_at = getattr(record, self.key)

        # if 'expires_at' is not set, that means it doesn't expire
        if expires_at is None:
            return False

        # comparing timezone-aware and naive datetimes results in an error
        # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive # noqa
        now = datetime.utcnow()
        if expires_at.tzinfo and expires_at.tzinfo.utcoffset(expires_at) is not None:
            now = now.replace(tzinfo=pytz.utc)

        return expires_at < now
